from ..record import Record, check_error, JsonData
from typing import Optional, ClassVar, Literal
import logging
import almapiwrapper.users as userslib
from almapiwrapper.inventory import Item

class Loan(Record):
    """Class representing a Users loan

    :cvar api_base_url_users: url of the user api
    :ivar loan_id: id of the fee
    :ivar item: :class:`almapiwrapper.inventory.Item` loaned item
    :ivar zone: initial value: zone of the fee
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar user: :class:`almapiwrapper.users.User` either primary_id of the user or the user itself must be provided
    :ivar data: :class:`almapiwrapper.record.JsonData` with fee data
    """

    api_base_url_users: ClassVar[str] = 'https://api-eu.hosted.exlibrisgroup.com/almaws/v1/users'

    def __init__(self,
                 loan_id: Optional[str] = None,
                 primary_id: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P',
                 user: Optional[userslib.User] = None,
                 data: Optional[JsonData] = None):
        """Constructor of Loan Object

        """
        # Fetch env and zone from user if available
        if user is not None:
            zone = user.zone
            env = user.env

        super().__init__(zone, env, data)
        self.area = 'Users'
        self.format = 'json'

        self._loan_id = loan_id
        self._item = None

        if primary_id is not None:
            self.user = userslib.User(primary_id, self.zone, self.env)
        elif user is not None:
            self.user = user
        else:
            logging.error('Missing information to construct a Loan')
            self.error = True

    def __repr__(self):
        """Get a string representation of the object. Useful for logs.
        :return: str
        """
        return f"{self.__class__.__name__}('{self.loan_id}', '{self.user.primary_id}', '{self.zone}', '{self.env}')"

    @property
    def loan_id(self) -> str:
        """Property returning the loan ID

        It fetches it in a private attribute if not available
        in the data property.
        """
        if self._data is not None:
            return self.data['loan_id']
        else:
            return self._loan_id

    @property
    def item(self) -> Optional[Item]:
        """Property returning the loaned item

        It fetches it in a private attribute if not available
        in the data property.
        """
        if self._item is None:
            item = self._fetch_item()
            if self.error is False:
                self._item = item

        return self._item

    def _fetch_data(self) -> Optional[JsonData]:
        """Fetch the json data of the loan

        :return: :class:`almapiwrapper.record.JsonData`"""
        r = self._api_call('get',
                           f'{self.api_base_url_users}/{self.user.primary_id}/loans/{self.loan_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: loan data available')
            fee_data = r.json()

            return JsonData(fee_data)

        else:
            self._handle_error(r, f'{repr(self)}: unable to fetch user loan')

    @check_error
    def _fetch_item(self) -> Item:
        """Fetch the loaned item

        :return: :class:`almapiwrapper.inventory.Item`"""
        return Item(self.data['mms_id'],
                    self.data['holding_id'],
                    self.data['item_id'],
                    self.zone,
                    self.env)

    @check_error
    def return_loan(self):
        pass

    @check_error
    def save(self) -> 'userslib.Loan':
        """Save a user loan record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/<primary_id>/loan_<IZ>_<loan_id>_<version>.xml

        :return: object :class:`almapiwrapper.users.Fee`
        """
        filepath = f'records/{self.user.primary_id}/loan_{self.zone}_{self.loan_id}.json'
        self._save_from_path(filepath)
        return self




"""This module allows getting information and changing Alma users"""

from typing import Optional, ClassVar, Literal, Union, List
import logging

import almapiwrapper.inventory

from ..record import Record, check_error, JsonData
import almapiwrapper.users as users
from datetime import datetime


class User(Record):
    """Class representing an Alma user

    :ivar primary_id: initial value: primary_id of the user
    :ivar zone: initial value: zone where the user should be created
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: initial value: :class:`almapiwrapper.record.JsonData` of the user (useful for new created users)
    """

    api_base_url: ClassVar[str] = 'https://api-eu.hosted.exlibrisgroup.com/almaws/v1/users'

    def __init__(self,
                 primary_id: str,
                 zone: str,
                 env: Optional[Literal['P', 'S']] = 'P',
                 data: Optional[JsonData] = None) -> None:
        """Constructor for user
        """
        super().__init__(zone, env, data)
        self.primary_id = primary_id
        self.area = 'Users'
        self.format = 'json'
        self._fees = None
        self._loans = None

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.
        :return: string
        """
        return f"{self.__class__.__name__}('{self.primary_id}', '{self.zone}', '{self.env}')"

    def _fetch_data(self) -> Optional[JsonData]:
        """Use API to fetch user data

        :return: json data or None if no data is available
        """

        r = self._api_call('get',
                           f'{self.api_base_url}/{self.primary_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: user data available')
            return JsonData(r.json())
        else:
            self._handle_error(r, f'unable to fetch user data')

    def _fetch_fees(self) -> Optional[List['users.Fee']]:
        """Fetch fee data of the current user

        :return: list of :class:`almapiwrapper.users.Fee` objects"""

        r = self._api_call('get',
                           f'{self.api_base_url}/{self.primary_id}/fees',
                           # params={'status': 'EXPORTED'},
                           headers=self._get_headers())
        if r.ok is True:

            logging.info(f'{repr(self)}: fees data available')
            fees_data = r.json()
            if 'fee' not in fees_data or fees_data['fee'] is None:
                logging.warning(f'{repr(self)}: no fee in the account')
                return []
            fees = []
            for fee_data in fees_data['fee']:
                fees.append(users.Fee(user=self, data=JsonData(fee_data)))
            return fees
        else:
            self._handle_error(r, f'unable to fetch user fees')

    def _fetch_loans(self) -> Optional[List['users.Loan']]:
        """Fetch loan data of the current user

        :return: list of :class:`almapiwrapper.users.Loan` objects"""

        r = self._api_call('get',
                           f'{self.api_base_url}/{self.primary_id}/loans',
                           params={'limit': '100'},
                           headers=self._get_headers())
        if r.ok is True:

            logging.info(f'{repr(self)}: loans data available')
            loans_data = r.json()
            if 'item_loan' not in loans_data or loans_data['item_loan'] is None:
                logging.warning(f'{repr(self)}: no loan in the account')
                return []
            loans = []
            for loan_data in loans_data['item_loan']:
                loans.append(users.Loan(user=self, data=JsonData(loan_data)))
            return loans
        else:
            self._handle_error(r, f'unable to fetch user loans')

    def save(self) -> 'User':
        """save() -> 'User'
        Save a user record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/<primary_id>/user_<IZ>_<primary_id>_<version>.xml

        :return: object :class:`almapiwrapper.users.User`
        """
        filepath = f'records/{self.primary_id}/user_{self.zone}_{self.primary_id}.json'
        self._save_from_path(filepath)
        return self

    @check_error
    def update(self, override: Optional[str] = None) -> 'User':
        """update() -> 'User'
        Update the user through api

        :param override: string containing the list of the fields to override

        :return: object :class:`almapiwrapper.users.User`

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        if override is not None:
            params = {'override': override}
        else:
            params = {}

        r = self._api_call('put',
                           f'{self.api_base_url}/{self.primary_id}',
                           data=bytes(self),
                           params=params,
                           headers=self._get_headers())
        if r.ok:
            logging.info(f'{repr(self)}: user updated.')
            self.data = JsonData(r.json())
        else:
            self._handle_error(r, 'failed to update user')

        return self

    @check_error
    def delete(self) -> None:
        """delete()
        Delete the user

        :return: None

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        r = self._api_call('delete',
                           f'{self.api_base_url}/{self.primary_id}',
                           headers=self._get_headers())
        if r.ok:
            logging.info(f'{repr(self)}: user deleted.')
        else:
            self._handle_error(r, 'failed to delete user')

    @property
    def primary_id(self) -> str:
        """Property returning the primary ID

        It fetches it in a private attribute if not available
        in the data property.

        When a new value is set and json data is available, it changes the primary ID
        at the data level.

        :return: primary ID of the user
        """
        if self._data is not None:
            return self.data['primary_id']
        else:
            return self._primary_id

    @primary_id.setter
    def primary_id(self, primary_id) -> None:
        """Set the primary ID

        If json data is available, change the primary ID
        at the data level.

        :param primary_id: new primary ID to set

        :return: None
        """
        self._primary_id = primary_id
        if self._data is not None:
            self.data['primary_id'] = self._primary_id

    @property
    def fees(self) -> Optional[List['users.Fee']]:
        """Property returning the list of the fees of the user

        :return: list of :class:`almapiwrapper.users.Fee` objects"""
        if self._fees is None:
            self._fees = self._fetch_fees()

        return self._fees

    @property
    def loans(self) -> Optional[List['users.Loan']]:
        """Property returning the list of the loans of the user

        :return: list of :class:`almapiwrapper.users.Loan` objects"""
        if self._loans is None:
            self._loans = self._fetch_loans()

        return self._loans

    @check_error
    def create_loan(self,
                    library: str,
                    circ_desk: str,
                    item_barcode: Optional[str] = None,
                    item_id: Optional[str] = None,
                    item: Optional[almapiwrapper.inventory.Item] = None) -> Optional['users.Loan']:
        """create_loan(library: str, circ_desk: str, item_barcode: Optional[str] = None, item_id: Optional[str] = None, item: Optional[almapiwrapper.inventory.Item] = None) -> Optional['users.Loan']

        Create a loan for the user

        :param library: code of the library
        :param circ_desk: code of the circulation desk
        :param item_barcode: barcode of the item
        :param item_id: id of the item
        :param item: object :class:`almapiwrapper.inventory.Item`

        :return: object :class:`almapiwrapper.users.Loan`

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        if item is not None and item.error is False:
            item_barcode = item.barcode

        # Check if mandatory fields are available
        if library is None or circ_desk is None or (item_barcode is None and item_id is None):
            logging.error('Missing information to create a loan')
            self.error = True
            return None

        # Create the loan
        loan_data = {"circ_desk": {"value": circ_desk},
                     "library": {"value": library}}

        params = {'item_barcode': item_barcode} if item_barcode is not None else {'item_pid': item_id}

        r = self._api_call('post',
                           f'{self.api_base_url}/{self.primary_id}/loans',
                           params=params,
                           data=bytes(JsonData(loan_data)),
                           headers=self._get_headers())
        if r.ok is True:
            loan_data = JsonData(r.json())

            loan = users.Loan(user=self, data=loan_data)

            logging.info(f'{repr(self)}: {repr(loan)} created.')

            return loan
        else:
            self._handle_error(r, f'unable to create loan for '
                                  f'{item_barcode if item_barcode is not None else item_id} '
                                  f'on {circ_desk} at {library}')

    @check_error
    def set_password(self, password: Optional[str] = '123pw123') -> 'User':
        """set_password(password: Optional[str] = '123pw123') -> 'User'
        Set a new password to a user

        :param password: password string, default is '123pw123'

        :return: object :class:`almapiwrapper.users.User`

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        self.data['password'] = password
        self.data['force_password_change'] = "TRUE"

        return self

    @check_error
    def add_synchro_note(self) -> 'User':
        """add_synchro_note() -> 'User'
        Add a test synchronization notes

        :return: object :class:`almapiwrapper.users.User`"""

        note = {"note_type": {"value": "OTHER"},
                "note_text": f"SLSP Test-synchronisation {datetime.now().strftime('%Y-%m-%d %H:%M')}",
                "user_viewable": False,
                "popup_note": False,
                "segment_type": "Internal"}
        self.data['user_note'].append(note)
        self.update()
        if self.error is False:
            logging.info(f'{repr(self)}: synchronization test note added')

        return self

    @check_error
    def remove_synchro_note(self) -> 'User':
        """remove_synchro_note() -> 'User'
        Remove all test synchronization notes

        :return: object :class:`almapiwrapper.users.User`"""
        nb_notes = len(self.data['user_note'])
        self.data['user_note'] = [note for note in self.data['user_note']
                                  if not note['note_text'].startswith('SLSP Test-synchronisation')]
        nb_sup_notes = nb_notes - len(self.data['user_note'])

        if nb_sup_notes > 0:
            self.update()
            if self.error is False:
                logging.info(f'{repr(self)}: {nb_sup_notes} synchronization test note(s) deleted')
        else:
            logging.warning(f'{repr(self)}: NO synchronization test note deleted')

        return self

    @check_error
    def check_synchro_note(self) -> bool:
        """check_synchro_note() -> bool
        Test if the user has a synchronization test note

        :return: True if a synchronization test note is found else False
        """
        has_synchro_note = len([note for note in self.data['user_note']
                                if note['note_text'].startswith('SLSP Test-synchronisation')]) > 0
        if has_synchro_note is True:
            logging.info(f'{repr(self)}: a synchronization test note exists')
        else:
            logging.warning(f'{repr(self)}: NO synchronization test note exists')
        return has_synchro_note


class NewUser(User):
    """Class used to create new users

    :var zone: initial value: zone where the user should be created
    :var env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :var data: initial value: :class:`almapiwrapper.record.JsonData` of the user to create
    """

    def __init__(self, zone: str, env: Literal['P', 'S'], data: JsonData):
        """Constructor to create new users
        """
        super(User, self).__init__(zone, env, data)
        self.area = 'Users'
        self.format = 'json'

    @check_error
    def create(self, password: Optional[str] = None) -> Union['NewUser', User]:
        """create(password: Optional[str] = None) -> Union['NewUser', User]
        Create the user with API.

        :param password: optional string with the password, if not provided,
            the password will be set at the default value (if not available in the data)

        :return: object :class:`almapiwrapper.users.User` or object :class:`almapiwrapper.users.NewUser`
            (in case of error)

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        self.set_password(password)

        r = self._api_call('post',
                           f'{self.api_base_url}',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok is True:
            logging.info(f'{repr(self)}: user created')
            self._data = JsonData(r.json())
            return User(self.primary_id, self.zone, self.env, self._data)
        else:
            self._handle_error(r, f'unable to create user')
            return self

    @check_error
    def set_password(self, password: Optional[str] = None) -> 'NewUser':
        """set_password(password: Optional[str] = None) -> 'NewUser'
        Set the password of a new user

        :param password: string containing the new password.

        :return: object :class:`almapiwrapper.users.NewUser`
        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        if password is None and self.data['password'] == '':
            password = '123pw123'

        if password is not None:
            self.data['password'] = password

        return self


if __name__ == "__main__":
    pass



import unittest

from almapiwrapper.users import User, NewUser, Loan
from almapiwrapper.record import JsonData
from almapiwrapper.inventory import Item
from almapiwrapper import config_log

config_log("test.log")


class TestCreateUser(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        for primary_id in ['testLoanUser1', 'testLoanUser2']:
            u = User(primary_id, 'UBS', 'S')

            if u.data is None:

                # Create new user
                data = JsonData(filepath=f'test/data/user_{primary_id}.json')
                _ = NewUser('UBS', 'S', data).create()

        Item('996259850105504', '22202377380005504', '23202377370005504', 'UBS', 'S').scan_in('A100', 'A100_KUR')

    def test_create_loan(self):

        u = User('testLoanUser1', 'UBS', 'S')

        # Create new loan
        loan = u.create_loan('A100', 'A100_KUR', '0UBU0536413')
        self.assertFalse(u.error, 'Error during loan creation')

        self.assertEqual(u.loans[0].data['loan_id'], loan.data['loan_id'], 'Loan ID mismatch')

    def test_scan_in(self):
        u = User('testLoanUser1', 'UBS', 'S')

        # Create new loan
        loan1 = u.create_loan('A100', 'A100_KUR', '0UBU0536413')
        self.assertEqual(loan1.data['loan_status'], 'ACTIVE', 'Loan status should be ACTIVE')

        item = loan1.item.scan_in('A100', 'A100_KUR')
        self.assertEqual(item.barcode, '0UBU0536413', 'Item status should be LOAN')

        loan2 = Loan(loan1.data['loan_id'], user=u)
        self.assertEqual(loan2.data['loan_status'], 'COMPLETE', 'Loan status should be COMPLETE')

    def test_loan_error(self):
        u1 = User('testLoanUser1', 'UBS', 'S')
        loan1 = u1.create_loan('A100', 'A100_KUR', '0UBU0536413')
        self.assertFalse(u1.error, 'Unexpected error during loan creation')
        self.assertIsNotNone(loan1, 'Loan should not be None')

        u2 = User('testLoanUser2', 'UBS', 'S')
        loan2 = u2.create_loan('A100', 'A100_KUR', '0UBU0536413')
        self.assertTrue(u2.error, 'No expected error during loan creation of already loaned item')
        self.assertIsNone(loan2, 'Loan should be None')

    @classmethod
    def tearDownClass(cls):
        for primary_id in ['testLoanUser1', 'testLoanUser2']:
            u = User(primary_id, 'UBS', 'S')
            if u.data is not None:
                for loan in u.loans:
                    loan.item.scan_in('A100', 'A100_KUR')
                u.delete()

if __name__ == '__main__':
    unittest.main()



import json
import os
from typing import Dict, Any, List, Optional, Literal


class ApiKeys:
    """This class help to manage multiple API Keys

    .. note:: The API keys should be stored in a json file. The path to
        this file has to be in `alma_api_keys` environment variable.

    :Example:

    >>> from almapiwrapper import ApiKeys
    >>> k = ApiKeys()
    >>> api_key_1 = k.get_key('NZ', 'Users', 'RW', 'P')

    Will get an API key for Network Zone with
    read and write rights on users
    in productive environment

    >>> api_key_2 = k.get_key('UBS', 'Bibs', 'RW', 'S')

    Will get an API key for UBS IZ with
    read and write rights on bibliographic data
    in productive environment.
    """
    def __init__(self) -> None:
        """
        Constructor: read the keys file and store it in the private attribute _keys.
        The path to the file should be in an environment variable "alma_api_keys"
        """
        self._keys = self._read_keys_file()

    @staticmethod
    def _read_keys_file() -> Dict[str, Any]:
        """
        Read the json file with API keys.
        :return: data with the api keys
        """
        if os.getenv('alma_api_keys') is None:
            return {}

        with open(os.getenv('alma_api_keys')) as f:
            return json.load(f)

    def get_key(self,
                zone: str,
                area: str,
                permissions: str,
                env: str = 'P') -> str:
        """Return the API key according to the requested parameters

        :param zone: abbreviated form of the IZ or of the NZ
        :param area: area of the API for example "Users"
        :param permissions: read or read/write. It accepts only "R" and "RW"
        :param env: production or sandbox environment, defaults to "P"
        :return: API key

        :raise: KeyError: exception is risen when no corresponding key is found
        """
        for k in self._keys[zone]:
            for api in k['Supported_APIs']:
                if api['Area'] == area and \
                    api['Permissions'] == permissions and \
                        api['Env'] == env:
                    return k['API_Key']
        raise KeyError(f'No corresponding API key found: zone "{zone}", area "{area}",'
                       f' permission "{permissions}", environment "{env}".')

    def get_iz_codes(self, env: Optional[Literal['P', 'S']] = 'P') -> List[str]:
        """
        Return the list of all IZ abbreviations
        :param env: "P" for production, "S" for sandbox. Default is production
        :return: Code of all IZ
        :rtype: str
        """
        if env == 'P':
            return [k for k in self._keys.keys() if k != 'NZ' or k.endswith('-Prod')]
        else:
            sb_izs = set()
            for zone in self._keys.keys():
                if zone != 'NZ':
                    for k in self._keys[zone]:
                        for api in k['Supported_APIs']:
                            if api['Env'] == 'S':
                                sb_izs.add(zone)
        return list(sb_izs)


if __name__ == "__main__":
    pass



import unittest
import sys
import os
import shutil

from almapiwrapper.inventory import IzBib, NzBib, Holding, Item
from almapiwrapper.record import JsonData, XmlData
from almapiwrapper import config_log
from lxml import etree

config_log("test.log")


class TestBib(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        IzBib('991043825829705501', 'UBS', 'S', from_nz_mms_id=True).delete()
        Item(barcode='03124510_NEW', zone='HPH', env='S').delete()

        _ = [hol.delete() for hol in IzBib('991000975799705520', zone='HPH', env='S').get_holdings()
             if hol.holding_id != '2234409340005520']

        item = Item(barcode='03124510_NEW_2', zone='HPH', env='S')
        if item.error is False:
            item.holding.delete(force=True)

        shutil.rmtree('records/UBS_9966145550105504', ignore_errors=True)
        item1 = Item(barcode='DSVN8068436', zone='UBS', env='S')

        if item1.data.find('.//internal_note_3').text is not None:
            item1.data.find('.//internal_note_3').text = None
            item1.update()

    def test_get_bib(self):

        # Get the data
        bib = IzBib('991000975799705520', 'HPH', 'S')
        title = bib.data.find('.//datafield[@tag="245"]/subfield[@code="a"]').text

        self.assertEqual(title, 'Je veux un python pour mon anniversaire',
                         'Title not corresponding to "Je veux un python pour mon anniversaire"')

        nz_mms_id = bib.get_nz_mms_id()

        self.assertEqual(nz_mms_id, '991043825829705501',
                         'NZ MMS ID not found (should be "991043825829705501")')

    def test_copy_from_nz(self):

        # Get the data from NZ id of a not existing record in IZ
        bib = IzBib('991043825829705501', 'UBS', 'S', from_nz_mms_id=True)

        self.assertTrue(bib.error,
                        'No error when fetching a not existing record in IZ from a NZ MMS ID: 991043825829705501')

        # Test copy record in IZ
        bib = IzBib('991043825829705501', 'UBS', 'S', from_nz_mms_id=True, copy_nz_rec=True)

        self.assertFalse(bib.error,
                         'Error when copying a record from a NZ  to IZ: MMS ID 991043825829705501')

        IzBib('991043825829705501', 'UBS', 'S', from_nz_mms_id=True).delete()

    def test_get_holdings_and_items(self):
        # Get holdings
        hols = IzBib('991043825829705501', 'HPH', 'S', from_nz_mms_id=True).get_holdings()

        self.assertGreater(len(hols), 1, "Error, not at least two holdings found")

        # Get items
        items = hols[0].get_items()

        self.assertTrue('03124510' in items[0].barcode, "Error, barcode item should start with '03124510'")

    def test_get_from_barcode(self):

        # Get the data from NZ id of a not existing record in IZ
        item = Item(barcode='NOT_EXISTING_BARCODE', zone='HPH', env='S')

        self.assertTrue(item.error,
                        'No error when fetching a not existing barcode in IZ')

        # Test with an existing barcode
        item = Item(barcode='03124510', zone='HPH', env='S')

        self.assertFalse(item.error,
                         'Error when fetching an item by barcode')

        self.assertEqual(item.holding.bib.mms_id, '991000975799705520',
                         'MMS ID of the bib record is not "991000975799705520"')

    def test_duplicate_item(self):

        # Get the item data
        item1 = Item(barcode='03124510', zone='HPH', env='S')

        # Change barcode
        item1.barcode = '03124510_NEW'

        item2 = Item(holding=item1.holding, data=item1.data, create_item=True)

        self.assertFalse(item2.error, 'No error when duplicating an item in IZ')
        self.assertEqual(item2.barcode, '03124510_NEW', 'Barcode of the duplicated item is not "03124510_NEW"')

        Item(barcode='03124510_NEW', zone='HPH', env='S').delete()

    def test_change_library_item(self):
        # Get the item data
        item1 = Item(barcode='03124510', zone='HPH', env='S')

        # Change barcode
        item1.barcode = '03124510_NEW_2'
        item1.library = 'hph_bjnju'
        item1.location = 'bjnjudoc'

        hol1 = item1.holding
        hol1.library = 'hph_bjnju'
        hol1.location = 'bjnjudoc'

        hol2 = Holding(bib=item1.bib, data=hol1.data, create_holding=True)
        item2 = Item(holding=hol2, data=item1.data, create_item=True)

        self.assertFalse(item2.error, 'No error when duplicating an item in IZ')
        self.assertEqual(item2.barcode, '03124510_NEW_2', 'Barcode of the duplicated item is not "03124510_NEW_2"')

        Item(barcode='03124510_NEW_2', zone='HPH', env='S').holding.delete(force=True)

    def test_holding_callnumber(self):
        holding = Holding('991000975799705520', '2234409340005520', 'HPH', 'S')
        self.assertEqual(holding.callnumber, 'HEPBE FRIE')

        holding.callnumber = 'Test'

        self.assertEqual(holding.callnumber, 'Test')

    def test_create_update_holding(self):
        # Get the holding data
        holding1 = Holding('991000975799705520', '2234409340005520', 'HPH', 'S')

        # Duplicate holding
        holding2 = Holding(bib=holding1.bib, data=holding1.data, create_holding=True)

        self.assertEqual(holding2.data.find('.//datafield[@tag="852"]/subfield[@code="b"]').text, 'hph_bjnbe')

        # Duplicate holding
        holding2.data.find('.//record').append(etree.XML(
            '<datafield tag="035" ind1=" " ind2=" "><subfield code="a">400040004000</subfield></datafield>'))

        holding2.update()

        holding3 = Holding(holding2.bib.mms_id, holding2.holding_id, 'HPH', 'S')

        self.assertEqual(holding3.data.findall('.//datafield[@tag="035"]/subfield[@code="a"]')[-1].text, '400040004000')

        holding2.delete()

    def test_update_bib_record(self):

        # Get some bib record
        bib = NzBib('991043825829705501', 'S')

        # Change the title
        bib.data.find('.//datafield[@tag="245"]/subfield[@code="a"]') \
                .text = 'Je ne veux plus de python pour mon anniversaire'

        bib.update()

        bib = NzBib('991043825829705501', 'S')
        title = bib.data.find('.//datafield[@tag="245"]/subfield[@code="a"]').text

        self.assertFalse(bib.error, 'No error when updating an item in NZ')
        self.assertEqual(title,
                         'Je ne veux plus de python pour mon anniversaire',
                         'Le titre ne correspond pas à la modification')

        # Load the original record from disk to restore the title
        data = XmlData(filepath='test/data/nz_record_orig.xml')
        bib = NzBib('991043825829705501', 'S', data=data)
        bib.update()

        # Check if the record have been updated
        bib = NzBib('991043825829705501', 'S')
        title = bib.data.find('.//datafield[@tag="245"]/subfield[@code="a"]').text

        self.assertFalse(bib.error, 'No error when updating an item in NZ')
        self.assertEqual(title,
                         'Je veux un python pour mon anniversaire',
                         'Original title not restored')

    def test_create_bib_record(self):
        bib = NzBib('991060046869705501', 'S')

        new_bib = NzBib(data=bib.data, env='S', create_bib=True)

        self.assertNotEqual(bib.mms_id, new_bib.mms_id, 'MMS ID of new record must be different of source record')

        new_bib.delete()
        self.assertFalse(new_bib.error, 'No error when deleting record in NZ')

    def test_get_records_id(self):
        item = Item(barcode='03124510', zone='HPH', env='S')

        self.assertEqual(item.get_nz_mms_id(),
                         item.bib.get_nz_mms_id(),
                         'NZ MMS id different if fetched from item and from bib record')

        self.assertEqual(item.get_mms_id(),
                         item.bib.get_mms_id(),
                         'IZ MMS id different if fetched from item and from bib record')

        self.assertEqual(item.get_holding_id(),
                         item.holding.get_holding_id(),
                         'Holding id different if fetched from item and from holding record')

    def test_get_from_disk(self):
        item1 = Item(barcode='DSVN8068436', zone='UBS', env='S')
        item1.save()
        item1.data.find('.//internal_note_3').text = 'changed on disk'
        item1.save()

        data_item2 = Item.get_data_from_disk(item1.get_mms_id(), item1.get_holding_id(), item1.get_item_id(), 'UBS')
        item2 = Item(item1.get_mms_id(),
                     item1.get_holding_id(),
                     item1.get_item_id(),
                     zone='UBS', env='S', data=data_item2)
        self.assertEqual(item2.data.find('.//internal_note_3').text,
                         'changed on disk',
                         'Internal note present on disk data')
        item2.update()

        self.assertFalse(item2.error, 'No error when updating record from disk (1)')

        data_item3 = XmlData(filepath='records/UBS_9966145550105504/item_22224723340005504_23224723290005504_01.xml')

        item3 = Item(item1.get_mms_id(),
                     item1.get_holding_id(),
                     item1.get_item_id(),
                     zone='UBS', env='S',
                     data=data_item3)

        item3.update()
        self.assertFalse(item3.error, 'No error when updating record from disk (2)')

        self.assertIsNone(item3.data.find('.//internal_note_3').text,
                          'Internal note removed from item')

        self.assertIsNone(Holding.get_data_from_disk(item1.get_mms_id(),
                                                     item1.holding.get_holding_id(),
                                                     'UBS'),
                          'No holding data saved on the disk')

        self.assertIsNone(IzBib.get_data_from_disk(item1.get_mms_id(), 'UBS'),
                          'No bib data saved on the disk')

        item1.holding.save()
        item1.bib.save()

        self.assertIsNotNone(Holding.get_data_from_disk(item1.get_mms_id(),
                                                        item1.holding.get_holding_id(),
                                                        'UBS'),
                             'Holding data available on the disk')

        self.assertIsNotNone(IzBib.get_data_from_disk(item1.get_mms_id(), 'UBS'),
                             'Bib data available on the disk')

    @classmethod
    def tearDownClass(cls):
        IzBib('991043825829705501', 'UBS', 'S', from_nz_mms_id=True).delete()
        Item(barcode='03124510_NEW', zone='HPH', env='S').delete()
        item = Item(barcode='03124510_NEW_2', zone='HPH', env='S')
        if item.error is False:
            item.holding.delete(force=True)

        _ = [hol.delete() for hol in IzBib('991000975799705520', zone='HPH', env='S').get_holdings()
             if hol.holding_id != '2234409340005520']


if __name__ == '__main__':
    unittest.main()



"""This module allow to create and manage sets in Alma"""
from typing import Literal, Optional, Union
import logging
from ..record import XmlData, Record, check_error
from ..inventory import IzBib, NzBib
from ..users import User
from lxml import etree


class RecSet(Record):
    """Class representing a set of records

    To get a set it is possible to provide either the set id or the name of the set.

    :ivar set_id: initial value: ID of the set
    :ivar zone: initial value: zone of the fee
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar name: initial value: name of the set
    """
    def __init__(self,
                 set_id: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Literal['P', 'S'] = 'P',
                 name: Optional[str] = None,
                 data: Optional[XmlData] = None) -> None:
        """Constructor of `RecSet`
        """
        super().__init__(zone, env, data)
        self.area = 'Conf'
        self.format = 'xml'
        self._name = name
        self.set_id = set_id
        self._members = None

    def _fetch_data(self) -> Optional[XmlData]:
        """This method fetch the data describing the set.

        It is possible to fetch the data of a set by providing the set_id or the name of the set.

        :return: :class:`almapiwrapper.record.XmlData`
        """
        # set_id is not provided. We try to fetch it from the name.
        if self.set_id is None and self._name is not None:
            r = self._api_call('get',
                               f'{self.api_base_url}/conf/sets',
                               params={'q': f'name~{self._name}'},
                               headers=self._get_headers())
            if r.ok is True and XmlData(r.content).content.find('.//set') is not None:
                logging.info(f'{repr(self)}: set data available')
                set_data = XmlData(r.content).content.find('.//set')

                # We store the set_id for fetching the complete data
                self.set_id = set_data.find('.//id').text

            else:
                logging.error(f'{repr(self)}: unable to fetch set data')
                self.error = True
                return None

        # set_id is provided. We fetch the data.
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/sets/{self.set_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: set data available')
            return XmlData(r.content)
        else:
            self._handle_error(r, 'unable to fetch set data')

    @property
    def name(self) -> str:
        """Property returning the name of the set.

        :return: str containing set name
        """
        if self._name is None:
            name = self.data.find('.//name')
            if name is not None:
                self._name = name.text

        return self._name

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        if self.set_id is not None:
            return f"{self.__class__.__name__}('{self.set_id}', '{self.zone}', '{self.env}')"
        else:
            return f"{self.__class__.__name__}(name='{self.name}', zone='{self.zone}', env='{self.env}')"

    @check_error
    def get_members_number(self) -> int:
        """get_members_number(self) -> int
        Return the number of members

        :return: int with the number of records

        .. note::
            If the record encountered an error, this
            method will be skipped."""
        return int(self.data.find('.//number_of_members').text)

    @check_error
    def get_set_type(self) -> str:
        """get_set_type(self) -> str
        Return the type of the set

        :return: str indicating the type of the set

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        return self.data.find('.//type').text

    @check_error
    def get_content_type(self) -> str:
        """get_content_type(self) -> str
        Return the type of the set

        :return: str indicating the type of the set

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        return self.data.find('.//content').text

    @check_error
    def delete(self) -> None:
        """delete(self) -> None
        Delete a set

        :return: None

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """

        r = self._api_call('delete',
                           f'{self.api_base_url}/conf/sets/{self.set_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: set deleted')
            return
        else:
            print(r.text)
            self._handle_error(r, f'unable to delete the set')

    @check_error
    def get_members(self) -> list[Record]:
        """get_members(self) -> list[Record]
        Return a list with all the records of the set

        :return: list of records depending on the type of the set

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """

        # Prevent to fetch data twice, if already available
        if self._members is not None:
            return self._members

        self._members = []
        while len(self._members) < self.get_members_number():
            r = self._api_call('get',
                               f'{self.api_base_url}/conf/sets/{self.set_id}/members',
                               params={'limit': '100', 'offset': str(len(self._members))},
                               headers=self._get_headers())
            if r.ok is False:
                self._handle_error(r, f'{repr(self)}: unable to fetch set members')
                return self._members
            else:
                rec_ids = [rec_id.text for rec_id in XmlData(r.content).content.findall('.//member/id')]

                # Build members -> IZ mms_id
                if self.get_content_type() in ['BIB_MMS', 'IEP'] and self.zone != 'NZ':
                    self._members += [IzBib(rec_id, self.zone, self.env) for rec_id in rec_ids]

                # Build members -> IZ mms_id
                elif self.get_content_type() in ['BIB_MMS'] and self.zone == 'NZ':
                    self._members += [NzBib(rec_id, self.env) for rec_id in rec_ids]

                # Build members -> users
                elif self.get_content_type() == 'USER':
                    self._members += [User(rec_id, self.zone, self.env) for rec_id in rec_ids]

                # Type of content not integrated
                else:
                    self._members += rec_ids
                    logging.warning(f'{repr(self)}: "{self.get_content_type()}" not integrated, only id are returned.')

                logging.info(f'{repr(self)}: {len(self._members)} / {self.get_members_number()} members fetched')

        return self._members

class NewRecSet(RecSet):
    """Class representing a set of records

    To create a new set, it is possible to provide directly the XML data or
    to give all the required parameters.

    :ivar zone: initial value: zone of the set
    :ivar env: initial value: environment of the set: 'P' for production and 'S' for sandbox

    :param name: name of the set
    :param description: description of the set
    :param query: query to build the set
    :param created_by: user who created the set
    :param private: set as private or not
    :param data: :class:`almapiwrapper.record.XmlData` object, can be used instead of the XML data of the set
    """
    def __init__(self,
                 zone: str,
                 env: Literal['P', 'S'] = 'P',
                 name: Optional[str] = None,
                 description: Optional[str] = None,
                 query: Optional[str] = None,
                 created_by: Optional[str] = None,
                 private: Optional[bool] = False,
                 data: Optional[XmlData] = None) -> None:
        """Constructor of `NewRecSet`
        """
        super(RecSet, self).__init__(zone, env, data)
        self.area = 'Conf'
        self.format = 'xml'
        self._members = None
        if self._data is None and name is not None and query is not None and created_by is not None:
            self._data = self._create_data(name, query, description, created_by, private)
        else:
            logging.error('NewRecSet: missing parameters to create a new set')

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.zone}', '{self.env}', '{self.data.find('name').text}', " \
               f"'{self.data.find('description').text}', '{self.data.find('query').text}', " \
               f"'{self.data.find('created_by').text}', {self.data.find('private').text=='true'})"

    @staticmethod
    def _create_data(name, query, description, created_by, private):
        """Create the XML data for a new set

        :param name: name of the set
        :param query: query of the set
        :param description: description of the set
        :param created_by: user who created the set
        :param private: bool indicating if set is private or public
        :return: XML data
        """
        content = query.split()[0]
        root = etree.Element('set')
        etree.SubElement(root, 'name').text = name
        etree.SubElement(root, 'type').text = 'LOGICAL'
        etree.SubElement(root, 'content').text = content
        etree.SubElement(root, 'description').text = description
        etree.SubElement(root, 'query').text = query
        etree.SubElement(root, 'created_by').text = created_by
        etree.SubElement(root, 'private').text = str(private).lower()
        return XmlData(etree.tostring(root))

    @check_error
    def create(self) -> Union['RecSet', 'NewRecSet']    :
        """Create the set

        :return: `RecSet`
        """
        r = self._api_call('post',
                           f'{self.api_base_url}/conf/sets',
                           headers=self._get_headers(),
                           data=bytes(self))
        if r.ok is True:
            logging.info(f'{repr(self)}: set created')
            self._data = XmlData(r.content)
            self.set_id = self.data.find('.//id').text
            return RecSet(self.set_id, self.zone, self.env, data=self._data)
        else:
            self._handle_error(r, f'unable to create the set')
            return self



"""This module allows getting information and changing Alma item records"""
import os
from typing import Optional, ClassVar, Literal
import logging
from ..record import Record, check_error, XmlData
import almapiwrapper.inventory as inventory
from lxml import etree
import re


class Item(Record):
    """Class representing an item.

    There are several possibilities to get item information:
        - 'get_items' method of :class:`almapiwrapper.inventory.Holding` object
        - Item(mms_id, holding_id, item_id, zone, env)
        - Item(holding=:class:`almapiwrapper.inventory.Holding`, item_id=item_id)
        - Item(barcode='barcode', zone='zone', env='env')

    Create an item:
        - Item(holding=:class:`almapiwrapper.inventory.Holding`, data=`almapiwrapper.record.XmlData`, create_item=True)

    :ivar item_id: numerical ID of the item
    :ivar zone: zone of the record
    :ivar env: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar holding: :class:`almapiwrapper.inventory.Holding` object
    :ivar data: :class:`almapiwrapper.record.XmlData` object, useful to force
        update a record from a backup
    :param barcode: string with barcode of the item
    :param mms_id: bib record mms_id
    :param holding_id: holding record ID
    :param create_item: when True, create a new item
    """

    api_base_url_items: ClassVar[str] = f'{Record.api_base_url}/items'
    api_base_url_bibs: ClassVar[str] = f'{Record.api_base_url}/bibs'

    def __init__(self,
                 mms_id: Optional[str] = None,
                 holding_id: Optional[str] = None,
                 item_id: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P',
                 holding: Optional[inventory.Holding] = None,
                 barcode: Optional[str] = None,
                 data: Optional[etree.Element] = None,
                 create_item: Optional[bool] = False) -> None:
        """Construct an item.
        """
        self.error = False
        self.error_msg = None
        self._data = data

        if holding_id is not None and mms_id is not None and holding is None:
            holding = inventory.Holding(mms_id, holding_id, zone, env)

        self.holding = holding
        self.item_id = item_id
        self._barcode = barcode
        self.area = 'Bibs'
        self.format = 'xml'

        # Set 'env' and 'zone' either from holding data either from parameters
        if holding is not None:
            self.zone = holding.zone
            self.env = holding.env
        else:
            self.zone = zone
            self.env = env

        # Create a new item if 'create_item' parameter is True
        if self.item_id is None and data is not None and holding is not None and create_item is True:
            self._create_item(data)

        # Fetch the item from the barcode, zone and env must be available
        elif barcode is not None and create_item is False and data is None:
            data = self._fetch_data(barcode)
            if data is not None:
                self.data = data
                self.item_id = self.get_item_id()

        # Fetch the item from IDs of item, holding and bibliographic record
        elif self.item_id is not None and mms_id is not None and holding_id is not None and data is None:
            self.holding = inventory.Holding(mms_id, holding_id, self.zone, self.env)
            if self.holding.error is True:
                self.error = True
            self.data = self._fetch_data()

        # Fetch item through 'item_id' and holding
        elif self.item_id is not None and holding is not None and data is None:
            self.data = self._fetch_data()

        # The data provided does not allow to initialize an item
        elif data is None:
            logging.error('Missing information to construct an Item')
            self.error = True

    def __repr__(self):
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        if self._holding is None and self._barcode is not None:
            return f"{self.__class__.__name__}(barcode='{self._barcode}', " \
                   f"zone='{self.zone}', env='{self.env}')"
        return f"{self.__class__.__name__}('{self.bib.mms_id}', '{self.holding.holding_id}'," \
               f"'{self.item_id}', '{self.zone}', '{self.env}')"

    def _fetch_data(self, barcode: Optional[str] = None) -> Optional[XmlData]:
        """
        Fetch item data and store it in 'data' attribute
        :param barcode: barcode of the item

        :return: None
        """
        if barcode is not None:
            r = self._api_call('get',
                               self.api_base_url_items,
                               params={'item_barcode': barcode},
                               headers=self._get_headers())
            if r.ok is True:
                logging.info(f"{repr(self)}: item data fetched with barcode '{barcode}'")
                return XmlData(r.content)
            else:
                self._handle_error(r, f"unable to get item data from barcode '{barcode}'")
                self.error = True
                return

            # # Fetch holding
            # self.holding = inventory.Holding(self.data.find('.//mms_id').text,
            #                                  self.data.find('.//holding_id').text,
            #                                  self.zone, self.env)

        # No barcode provided
        else:
            r = self._api_call('get',
                               f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding.holding_id}'
                               f'/items/{self.item_id}',
                               headers=self._get_headers())

            if r.ok is True:

                logging.info(f'{repr(self)}: item data available')
                return XmlData(r.content)
            else:
                self._handle_error(r, 'unable to fetch item data')

    def _create_item(self, data: etree.Element):
        """
        Create an item to the holding with the provided data.
        :param data: item data

        :return: Item
        """
        r = self._api_call('post',
                           f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding.holding_id}/items',
                           headers=self._get_headers(),
                           data=etree.tostring(data))

        if r.ok is True:
            self.data = XmlData(r.content)
            self.item_id = self.get_item_id()
            logging.info(f'{repr(self)}: item created')
        else:
            self._handle_error(r, 'unable to create the new item')

    @check_error
    def get_item_id(self) -> str:
        """get_item_id() -> str
        Fetch the item ID in the xml data of 'data' attribute. Useful for creating a new item.

        :return: item id
        """
        return self.data.find('.//pid').text

    @check_error
    def get_holding_id(self) -> str:
        """get_holding_id() -> str:
        Fetch holding ID

        :return: holding ID
        """

        return self.data.find('.//holding_id').text

    @check_error
    def get_mms_id(self) -> str:
        """get_mms_id(self) -> str
        Fetch IZ MMS ID

        :return: IZ MMS ID
        """

        return self.data.find('.//mms_id').text

    @check_error
    def get_nz_mms_id(self) -> Optional[str]:
        """get_mms_id(self) -> Optional[str]
        Fetch NZ MMS ID

        :return: NZ MMS ID
        """

        nz_number_fields = self.data.findall('.//network_number')
        for field in nz_number_fields:
            m = re.match(r'\(EXLNZ.+?\)(\d+)', field.text)
            if m is not None:
                return m.group(1)

        logging.warning('{repr(self)}: Record not linked with NZ, no NZ MMS ID available')

    @property
    def bib(self) -> inventory.IzBib:
        """
        Property of the item returning the bibliographic record

        :return: IzBib
        """
        if self.holding is not None:
            return self.holding.bib

    @property
    def holding(self) -> Optional[inventory.Holding]:
        """holding(self) -> Optional[inventory.Holding]
        Property of the item returning the :class:`almapiwrapper.inventory.Holding` object
        related to the item

        :return: :class:`almapiwrapper.inventory.Holding`
        """
        if self._holding is None and self.error is False:
            holding_id = self.get_holding_id()
            mms_id = self.get_mms_id()
            self._holding = inventory.Holding(mms_id, holding_id, self.zone, self.env)

        # To force fetching holding data
        # _ = self._holding.data

        if self._holding.error is True:
            self.error = True
            return

        return self._holding

    @holding.setter
    def holding(self, holding: inventory.Holding) -> None:
        """Property of the item containing the holding

        :return: None"""
        self._holding = holding

    @property
    @check_error
    def barcode(self) -> Optional[str]:
        """barcode(self) -> Optional[str]
        Property of the item returning the barcode

        :return: library code
        """
        barcode_field = self.data.find('.//barcode')
        if barcode_field is None:
            logging.warning(f'{repr(self)}: no barcode in the item')
            return
        return barcode_field.text

    @barcode.setter
    @check_error
    def barcode(self, barcode: str) -> None:
        """barcode(self, barcode: str) -> None
        This setter is able to update the barcode of the item. But the field should already exist.

        :param barcode: barcode of the item

        :return: None
        """
        barcode_field = self.data.find('.//barcode')
        if barcode_field is None:
            logging.error(f'{repr(self)}: no barcode field in the item -> not possible to update it')
            self.error = True
        else:
            logging.info(f'{repr(self)}: barcode changed from "{barcode_field.text}" to "{barcode}"')
            barcode_field.text = barcode

    @check_error
    def save(self) -> 'Item':
        """save(self) -> 'Item'
        Save a record item in the 'records' folder. Versioning is supported. A suffix is added to the file path.
        Example: records/UBS_9963486250105504/item_22314215800005504_23314215790005504_01.xml

        :return: Item
        """
        filepath = f'records/{self.zone}_{self.bib.mms_id}/item_{self.holding.holding_id}_{self.item_id}.xml'
        self._save_from_path(filepath)
        return self

    @check_error
    def update(self) -> 'Item':
        """update(self) -> 'Item'
        Update items data.

        :return: Item
        """
        r = self._api_call('put',
                           f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding.holding_id}/'
                           f'items/{self.item_id}',
                           data=bytes(self),
                           headers=self._get_headers())

        if r.ok is True:
            self.data = XmlData(r.content)
            logging.info(f'{repr(self)}: item data updated')
        else:
            self._handle_error(r, 'unable to update item data')

        return self

    @check_error
    def delete(self) -> None:
        """delete(self) -> None
        Delete an item.

        :return: None
        """
        r = self._api_call('delete',
                           f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding.holding_id}'
                           f'/items/{self.item_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)} deleted')
        else:
            self._handle_error(r, 'unable to delete the item')

    @property
    @check_error
    def library(self) -> Optional[str]:
        """library(self) -> Optional[str]
        Property of the holding returning the library code

        :return: library code
        """
        library = self.data.find('.//item_data/library')
        if library is None:
            logging.warning(f'{repr(self)}: no library in the item')
            return
        return library.text

    @library.setter
    @check_error
    def library(self, library_code: str) -> None:
        """library(self, library_code: str) -> None
        This setter is able to update the 852$b of the holding. But the field should already exist.

        :param library_code: code of the library to insert in 852$b field

        :return: None
        """
        library = self.data.find('.//item_data/library')
        if library is None:
            logging.error(f'{repr(self)}: no library field -> not possible to update it')
            self.error = True
        else:
            logging.info(f'{repr(self)}: library changed from "{library.text}" to "{library_code}"')
            library.text = library_code
            library.attrib.pop('desc', None)

    @property
    @check_error
    def location(self) -> Optional[str]:
        """location(self) -> Optional[str]
        Property of the holding returning the library code

        :return: library code
        """
        location = self.data.find('.//item_data/location')
        if location is None:
            logging.warning(f'{repr(self)}: no location in the item')
            return
        return location.text

    @location.setter
    @check_error
    def location(self, location_code: str) -> None:
        """location(self, location_code: str) -> None
        This setter is able to update the 852$c of the holding. But the field should already exist.

        :param location_code:

        :return: None
        """
        location = self.data.find('.//item_data/location')
        if location is None:
            logging.error(f'{repr(self)}: no location field -> not possible to update it')
        else:
            logging.info(f'{repr(self)}: location changed from "{location.text}" to "{location_code}"')
            location.text = location_code
            location.attrib.pop('desc', None)

    @staticmethod
    def get_data_from_disk(mms_id: str, holding_id: str, item_id: str, zone: str) -> Optional[XmlData]:
        """get_data_from_disk(mms_id, holding_id, item_id, zone)
        Fetch the data of the described record

        :param mms_id: bib record mms_id
        :param holding_id: holding record ID
        :param item_id: numerical ID of the item
        :param zone: zone of the record

        :return: :class:`almapiwrapper.record.XmlData` or None
        """
        if os.path.isdir(f'records/{zone}_{mms_id}') is False:
            return

        # Fetch all available filenames related to this record
        file_names = sorted([file_name for file_name in os.listdir(f'records/{zone}_{mms_id}')
                             if file_name.startswith(f'item_{holding_id}_{item_id}') is True])

        if len(file_names) == 0:
            return

        return XmlData(filepath=f'records/{zone}_{mms_id}/{file_names[-1]}')

    @check_error
    def scan_in(self, library: str, circ_desk: str) -> 'Item':
        """scan_in(self) -> 'Item'
        Scan in an item

        :return: Item
        """
        r = self._api_call('post',
                           f'{self.api_base_url_bibs}/{self.bib.mms_id}/holdings/{self.holding.holding_id}/'
                           f'items/{self.item_id}',
                           params={'op': 'scan', 'library': library, 'circ_desk': circ_desk},
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: scanned in')
        else:
            self._handle_error(r, 'unable to "scan in" the item')

        return self



"""This module allows getting information and changing Alma bib records"""

from typing import Optional, ClassVar, Literal, Union, List
import abc
import logging
from ..record import Record, check_error, XmlData
from lxml import etree
from copy import deepcopy
import almapiwrapper.inventory as inventory
import os


class Bib(Record, metaclass=abc.ABCMeta):
    """Class representing bibliographic record

    This abstract class groups common methods to "IzBib" and
    "NzBib". Only these two classes should be instanced.

    :ivar mms_id: record mms_id
    :ivar zone: zone of the record
    :ivar env: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: :class:`almapiwrapper.record.XmlData`
        object, useful to force update a record from a backup
    """

    api_base_url_bibs: ClassVar[str] = f'{Record.api_base_url}/bibs'

    def __init__(self,
                 mms_id: str,
                 zone: str,
                 env: Literal['P', 'S'] = 'P',
                 data: Optional[XmlData] = None) -> None:
        """Constructor for bib records

        Generic constructor for IZ and NZ bibliographic records. This method is called
        by the "IzBib" and "NzBib" classes to initialize the common elements.
        """
        super().__init__(zone, env, data)
        self.mms_id = mms_id
        self.area = 'Bibs'
        self.format = 'xml'

    def _fetch_data(self) -> Optional[XmlData]:
        """Fetch bibliographic data and store it in the "data" attribute as an Etree element

        :return: None or :class:`almapiwrapper.record.XmlData` object
        """
        r = self._api_call('get',
                           f'{self.api_base_url_bibs}/{self.mms_id}',
                           headers=self._get_headers())

        if r.ok is True:
            logging.info(f'{repr(self)}: bib data available')
            return XmlData(r.content)
        else:
            self._handle_error(r, 'unable to fetch bib data')

    @check_error
    def get_mms_id(self) -> str:
        """get_mms_id(self) -> str
        Fetch the MMS ID in controlfield 001

        Useful when fetching the bibliographic record with the NZ ID.

        :return: string with MMS ID of 001 controlfield
        """
        return self.data.find('.//controlfield[@tag="001"]').text

    @check_error
    def sort_fields(self) -> 'Bib':
        """sort_fields(self) -> 'Bib'
        Sort all the fields and subfields of the record

        :return: Bib
        """
        if self._data is not None:
            self._data.sort_fields()
        return self

    @check_error
    def update(self) -> 'Bib':
        """update(self) -> 'Bib'
        Update data

        On BibIz records this method is used to change local fields. On BibNz records, other fields can
        be changed.

        :return: Bib
        """
        r = self._api_call('put',
                           f'{self.api_base_url_bibs}/{self.mms_id}',
                           data=bytes(self),
                           headers=self._get_headers())

        if r.ok is True:
            self.data = XmlData(r.content)
            logging.info(f'{repr(self)}: bib data updated in {self.zone}')
        else:
            print(r.text)
            self._handle_error(r, 'unable to update bib data')

        return self

    @check_error
    def save(self) -> 'Bib':
        """save(self) -> 'Bib'
        Save a record in the 'records' folder

        Versioning is supported. A suffix is added to the file path.

        Example: records/NZ_991170519490005501/bib_991170519490005501_01.xml

        :return: Bib
        """
        filepath = f'records/{self.zone}_{self.mms_id}/bib_{self.mms_id}.xml'
        self._save_from_path(filepath)
        return self

    @check_error
    def add_fields(self, fields: Union[etree.Element, List[etree.Element]]) -> 'Bib':
        """add_fields(self, fields: Union[etree.Element, List[etree.Element]]) -> 'Bib'
        Add fields to the data of the current record

        :param fields: must be an etree element or a list of etree elements
        :return: Bib
        """
        # If fields is only one etree element, then transform it to a list
        if type(fields) is not list:
            fields = [fields]
        record = self.data.find('.//record')

        if record is None:
            # No record found
            logging.error(f'{repr(self)}: adding fields failed, no record available in data attribute')
            self.error = True
            return self

        # Add the fields
        for field in fields:
            record.append(deepcopy(field))
        logging.info(f'{repr(self)}: {len(fields)} fields added to the record')

        # Sort fields
        _ = self.sort_fields()

        return self

    @staticmethod
    def get_data_from_disk(mms_id: str, zone: str) -> Optional[XmlData]:
        """get_data_from_disk(mms_id, holding_id, item_id, zone)
        Fetch the data of the described record

        :param mms_id: bib record mms_id
        :param zone: zone of the record

        :return: :class:`almapiwrapper.record.XmlData` or None
        """
        if os.path.isdir(f'records/{zone}_{mms_id}') is False:
            return

        # Fetch all available filenames related to this record
        file_names = sorted([file_name for file_name in os.listdir(f'records/{zone}_{mms_id}')
                             if file_name.startswith(f'bib_{mms_id}') is True])

        if len(file_names) == 0:
            return

        return XmlData(filepath=f'records/{zone}_{mms_id}/{file_names[-1]}')


class IzBib(Bib):
    """Class representing bibliographic record of the IZ

    It inherits from Bib for common methods with NzBib.

    :ivar mms_id: record mms_id
    :ivar zone: zone of the record
    :ivar env: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: :class:`almapiwrapper.record.XmlData` object, useful to force update a record from a backup
    :param from_nz_mms_id: if this parameter is True the system assumes that the provided MMS ID is a network ID
        and fetch data from it
    :param copy_nz_rec: if this parameter is True, if no record exists in the IZ for the provided
        NZ ID, the CZ record is copied from NZ

    """
    def __init__(self, mms_id: str,
                 zone: str,
                 env: Literal['P', 'S'] = 'P',
                 from_nz_mms_id: Optional[bool] = False,
                 copy_nz_rec: Optional[bool] = False,
                 data: Optional[XmlData] = None):
        """Constructor of an IZ bibliographic record
        """

        super().__init__(mms_id, zone, env, data)
        self._holdings = None
        if from_nz_mms_id is True:
            self.data = self._fetch_bib_data_from_nz_id(copy_nz_rec)
            self.mms_id = self.get_mms_id()

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.mms_id}', '{self.zone}', '{self.env}')"

    def _fetch_bib_data_from_nz_id(self, copy_nz_rec: Optional[bool] = False) -> Optional[XmlData]:
        """Check if the record exists already in the destination IZ

        :param copy_nz_rec: when True the record is copied from the NZ if it didn't already exist in the IZ
        :return: None or :class:`almapiwrapper.record.XmlData`
        """
        nz_mms_id = self.mms_id

        # Fetch data from nz mms_id
        r = self._api_call('get', f'{self.api_base_url_bibs}',
                           params={'nz_mms_id': nz_mms_id},
                           headers=self._get_headers())

        if r.ok is True:
            # Data found in the IZ for the NZ MMS ID provided
            logging.info(f'NZ {nz_mms_id}: bib data available in {self.zone} -> {repr(self)}')
            return XmlData(r.content)

        else:
            # Data not found in the IZ for the NZ MMS ID provided
            if copy_nz_rec is False:
                self._handle_error(r, 'unable to fetch bib data from NZ id')
            else:
                # if 'copy_nz_rec' parameter is True, it will try to copy the record from the NZ
                logging.warning(f'NZ MMS_ID {self.mms_id}: no data available in {self.zone}')
                return self._copy_record_from_nz()

    def _copy_record_from_nz(self) -> Optional[XmlData]:
        """Copy NZ record to IZ. Loads the data in 'data' attribute.
        :return: None
        """
        nz_mms_id = self.mms_id

        r = self._api_call('post', f'{self.api_base_url_bibs}',
                           params={'from_nz_mms_id': nz_mms_id},
                           data='<bib/>',
                           headers=self._get_headers())

        if r.ok is True:
            logging.info(f'Record {repr(self)} copied from NZ record {nz_mms_id}')
            return XmlData(r.content)
        else:
            self._handle_error(r, 'unable to copy NZ record to IZ')

    @check_error
    def get_nz_mms_id(self) -> Optional[str]:
        """get_nz_mms_id(self) -> Optional[str]
        Fetch the NZ MMS ID of the IZ bib record

        :return: string with NZ record MMS ID.
        """
        nz_mms_id = self.data.find('.//linked_record_id[@type="NZ"]')
        if nz_mms_id is not None:
            logging.info(f'{repr(self)}: get NZ mms_id: {nz_mms_id.text}')
            return nz_mms_id.text

        logging.error(f'{repr(self)}: no NZ MMS ID available')
        return None

    @check_error
    def delete(self, force: Optional[bool] = False) -> None:
        """delete(force: Optional[bool] = False) -> None
        Delete bibliographic record in the IZ

        To delete locally a record,
        it needs to be unlinked from the NZ and without holdings and items.

        :param force: when True delete also holdings and items
        :return: None
        """

        # Unlink NZ and IZ record
        r = self._api_call('post',
                           f'{self.api_base_url_bibs}/{self.mms_id}',
                           params={'op': 'unlink_from_nz'},
                           data='<bib/>',
                           headers=self._get_headers())

        # Delete record
        if r.ok is True:
            logging.info(f'{repr(self)} unlinked from NZ')

            r = self._api_call('delete',
                               f'{self.api_base_url_bibs}/{self.mms_id}',
                               headers=self._get_headers(), params={'override': 'true' if force is True else 'false'})
            if r.ok is True:
                logging.info(f'{repr(self)} deleted')
                return

        self._handle_error(r, 'unable to delete the record')

    @check_error
    def get_holdings(self) -> List['inventory.Holding']:
        """get_holdings(self) -> List['inventory.Holding']
        Get list of holdings and store it in '_holdings' attribute

        It avoids having to reload it.

        :return: list of :class:`almapiwrapper.inventory.Holding` objects
        """
        # Check if holdings already fetched
        if self._holdings is not None:
            return self._holdings

        r = self._api_call('get',
                           f'{self.api_base_url_bibs}/{self.mms_id}/holdings',
                           headers=self._get_headers())
        root = etree.fromstring(r.content, parser=self.parser)
        holdings_data = root.findall('.//holding')

        # No holding available
        if len(holdings_data) == 0:
            logging.warning(f'{repr(self)}: no holding found')
            self._holdings = []
            return self._holdings

        # List of holdings found
        logging.info(f'{repr(self)}: {len(holdings_data)} holdings fetched')

        self._holdings = []
        for holding in holdings_data:
            holding_id = holding.find('holding_id').text
            self._holdings.append(inventory.Holding(bib=self, holding_id=holding_id))

        return self._holdings

    @check_error
    def delete_holdings(self, force: Optional[bool] = False) -> None:
        """delete_holdings(self, force: Optional[bool] = False) -> None
        Delete all holdings of the record with items if 'force' is True.

        :param force: when True delete the items too.
        :return: None
        """
        for holding in self.get_holdings():
            holding.delete(force=force)

    @check_error
    def get_local_fields(self) -> List[etree.Element]:
        """get_local_fields(self) -> List[etree.Element]
        Fetch the list of the local fields of the records

        It looks for subfield "9" and then get the parent.

        :return: list of etree.Element
        """
        local_fields = [field.getparent() for field in
                        self.data.findall('.//record/datafield/subfield[@code="9"]')
                        if field.text.lower() == 'local']
        return local_fields


class NzBib(Bib):
    """Class representing a NZ bibliographic record.

    It inherits from Bib for common methods with IzBib.

    :ivar mms_id: record mms_id
    :ivar env: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: :class:`almapiwrapper.record.XmlData` object, useful to
        force update a record from a backup
    :param create_bib: bool, if True, create a new bib record in the NZ
    """
    def __init__(self, mms_id: Optional[str] = None,
                 env: Literal['P', 'S'] = 'P',
                 data: Optional[XmlData] = None,
                 create_bib: Optional[bool] = False) -> None:
        """
        Construct a bibliographic record of the NZ
        """
        if data is not None:
            if data.__class__.__name__ == '_Element':
                data = XmlData(etree.tostring(data))

        super().__init__(mms_id, 'NZ', env, data)

        if create_bib is True:
            self._create_bib()

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.
        :return: string
        """
        return f"{self.__class__.__name__}('{self.mms_id}', '{self.env}')"

    @check_error
    def delete(self, force: Optional[bool] = False) -> None:
        """delete(self) -> None
        Delete bibliographic record in the IZ

        To delete locally a record,
        it needs to be unlinked from the NZ and without holdings and items.

        :param force: when True delete also holdings and items
        :return: None
        """

        # Delete record
        r = self._api_call('delete',
                           f'{self.api_base_url_bibs}/{self.mms_id}',
                           params={'override': 'true' if force is True else 'false'},
                           headers=self._get_headers())

        if r.ok is True:
            logging.info(f'{repr(self)} deleted')
            return

        self._handle_error(r, 'unable to delete the NZ record')

    @check_error
    def _create_bib(self) -> None:
        """Create a new NZ bib record with API.

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """

        r = self._api_call('post',
                           f'{self.api_base_url_bibs}',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok is True:
            self.data = XmlData(r.content)
            self.mms_id = self.get_mms_id()
            logging.info(f'{repr(self)}: NZ bib record created')

        else:
            self._handle_error(r, f'unable to create NZ bib record')
